#![no_std]
#![no_main]

use cortex_m::delay::Delay;

use display_interface::DisplayError;

use embedded_graphics::{
    mono_font::{ascii::FONT_6X10, MonoTextStyleBuilder},
    pixelcolor::BinaryColor,
    prelude::*,
    text::{Baseline, Text},
};

use embedded_hal::digital::v2::{InputPin, ToggleableOutputPin};

use fugit::RateExtU32;

use panic_halt as _;

use rp_pico::{
    entry,
    hal::{
        clocks::{self, Clock},
        gpio::PinState,
        i2c::I2C,
        usb::UsbBus,
        Sio, Watchdog,
    },
    pac::{CorePeripherals, Peripherals},
    Pins,
};

use ssd1306::{prelude::*, I2CDisplayInterface, Ssd1306};

use usb_device::{bus::UsbBus as UsbBusTrait, bus::UsbBusAllocator, prelude::*};

use usbd_serial::{SerialPort, USB_CLASS_CDC};

fn print_serial<'a, B: UsbBusTrait>(
    serial: &mut SerialPort<'a, B>,
    msg: &str,
) -> Result<(), UsbError> {
    let buf = msg.as_bytes();
    let mut pos = 0;

    while pos < buf.len() {
        match serial.write(&buf[pos..]) {
            Ok(count) => {
                pos += count;
                continue;
            }
            Err(UsbError::WouldBlock) => continue,
            Err(err) => return Err(err),
        }
    }

    Ok(())
}

fn stringify_display_error(err: DisplayError) -> &'static str {
    match err {
        DisplayError::InvalidFormatError => "invalid format",
        DisplayError::BusWriteError => "bus write error",
        DisplayError::DCError => "dc error",
        DisplayError::CSError => "cs error",
        DisplayError::DataFormatNotImplemented => "data format not implemented",
        DisplayError::RSError => "rs error",
        DisplayError::OutOfBoundsError => "out of bounds",
        _ => "unknown error",
    }
}

#[entry]
fn main() -> ! {
    let mut p = Peripherals::take().unwrap();
    let cp = CorePeripherals::take().unwrap();

    let mut watchdog = Watchdog::new(p.WATCHDOG);
    let clocks = clocks::init_clocks_and_plls(
        rp_pico::XOSC_CRYSTAL_FREQ,
        p.XOSC,
        p.CLOCKS,
        p.PLL_SYS,
        p.PLL_USB,
        &mut p.RESETS,
        &mut watchdog,
    )
    .ok()
    .unwrap();

    let mut delay = Delay::new(cp.SYST, clocks.system_clock.freq().to_Hz());

    let sio = Sio::new(p.SIO);
    let pins = Pins::new(p.IO_BANK0, p.PADS_BANK0, sio.gpio_bank0, &mut p.RESETS);

    let usb_bus = UsbBusAllocator::new(UsbBus::new(
        p.USBCTRL_REGS,
        p.USBCTRL_DPRAM,
        clocks.usb_clock,
        true,
        &mut p.RESETS,
    ));

    let mut serial = SerialPort::new(&usb_bus);

    let mut usb_dev = UsbDeviceBuilder::new(&usb_bus, UsbVidPid(0x2342, 0x1337))
        .manufacturer("lujoga")
        .product("Soldering Iron")
        .device_class(USB_CLASS_CDC)
        .build();

    for _ in 0..2000 {
        delay.delay_ms(5);

        if !usb_dev.poll(&mut [&mut serial]) {
            continue;
        }

        let mut buf = [0u8];
        match serial.read(&mut buf[..]) {
            Ok(_) => break,
            _ => (),
        }
    }

    let _ = print_serial(&mut serial, "Hello, world!\r\n");

    let i2c = I2C::i2c0(
        p.I2C0,
        pins.gpio0.into_mode(),
        pins.gpio1.into_mode(),
        400.kHz(),
        &mut p.RESETS,
        clocks.peripheral_clock.freq(),
    );

    let mut display = Ssd1306::new(
        I2CDisplayInterface::new_custom_address(i2c, 0x3c),
        DisplaySize128x64,
        DisplayRotation::Rotate0,
    )
    .into_buffered_graphics_mode();

    let result = display.init().map_err(stringify_display_error);
    if let Err(msg) = result {
        let _ = print_serial(&mut serial, msg);
        panic!();
    }

    let text_style = MonoTextStyleBuilder::new()
        .font(&FONT_6X10)
        .text_color(BinaryColor::On)
        .build();

    let result = Text::with_baseline("Hello, world!", Point::zero(), text_style, Baseline::Top)
        .draw(&mut display)
        .map_err(stringify_display_error);
    if let Err(msg) = result {
        let _ = print_serial(&mut serial, msg);
        panic!();
    }

    let result = display.flush().map_err(stringify_display_error);
    if let Err(msg) = result {
        let _ = print_serial(&mut serial, msg);
    }

    let mut relay = pins.gpio14.into_push_pull_output_in_state(PinState::Low);
    let tmp_button = pins.gpio15.into_pull_up_input();

    let mut led = pins.led.into_push_pull_output_in_state(PinState::Low);

    let mut counter = 0;
    loop {
        delay.delay_ms(5);
        counter = (counter + 1) % 100;

        if counter == 0 {
            led.toggle().unwrap();

            let _ = print_serial(&mut serial, "ping\r\n");
        }

        if tmp_button.is_low().unwrap() {
            relay.toggle().unwrap();
            delay.delay_ms(1000);
        }

        if !usb_dev.poll(&mut [&mut serial]) {
            continue;
        }

        let mut buf = [0u8; 64];
        match serial.read(&mut buf[..]) {
            Ok(count) => {
                let _ = serial.write(&buf[..count]);
            }
            _ => (),
        }
    }
}
